from market_insights.core.column_definition import DerivativesBaseColumns, DerivativesCalculatedColumns
from market_insights.core.core import TypeHelper
import pandas as pd
import datetime
import os

class TradesResultBaseMetrics:
  pnl: float = 0
  capital_required: float = 0
  no_of_trades: int = 0
  non_tradables_no_entry_count: int = 0
  non_tradables_no_exit_count: int = 0
  turnover: float = 0

class TradesResultMetrics (TradesResultBaseMetrics):
  transaction_fees: float = 0
  brokerage: float = 0
  total_expense: float = 0
  net_pnl: float = 0
  roi: float = 0
  break_even_date: datetime.datetime = None

class TradesResultBase (TradesResultMetrics):
  summary: pd.DataFrame

  def __init__(self):
    self.summary = None

  def calculate(self):
    self.transaction_fees = round(self.turnover * 0.01)
    self.brokerage = round(self.no_of_trades * 20)
    self.total_expense = round(self.transaction_fees + self.no_of_trades)
    self.net_pnl = round(self.pnl - self.total_expense)
    if self.capital_required > 0:
      self.roi = round(self.net_pnl / self.capital_required * 100)

  def prepare_summary_df(self):
    self.calculate()
    summary_dict: dict = {}
   
    props = TypeHelper.get_class_props(TradesResultMetrics)

    for prop in props:
      summary_dict[prop] = getattr(self, prop)

    self.summary = pd.DataFrame(summary_dict, index=[0])
  
  def __str__(self):
     return f'Net PnL: {round(self.net_pnl)}, Capital Required: {round(self.capital_required)}, ROI: {round(self.roi)}%, Non tradables (No Entry): {self.non_tradables_no_entry_count}, Non tradables (No Exit): {self.non_tradables_no_exit_count}'

class TradesResult(TradesResultBase):
    trades: pd.DataFrame
    non_tradables: pd.DataFrame
    full_trades: pd.DataFrame
    
    def __init__(self, trades, full_trades, non_tradables):
      super().__init__()
      self.trades = trades
      self.non_tradables = non_tradables
      self.full_trades = full_trades

    def caclulate_from_df(self):
      if len(self.trades) > 0:
        self.trades = self.trades.sort_values(DerivativesBaseColumns.Date).reset_index(drop=True)
        self.trades['CumCredit'] = self.trades['Credit'].cumsum()
        self.capital_required = round(abs(self.trades['CumCredit'].min()))
        self.no_of_trades = self.trades[DerivativesBaseColumns.Identifier].count()
        self.pnl = round(self.full_trades['Net'].sum())
        self.turnover = round(self.full_trades['Turnover'].sum())
        self.break_even_date = self.get_break_even_date()

      super().calculate()
      
    def get_break_even_date(self):
      break_even_date = self.trades[self.trades['CumCredit'] <= 0].tail(1)[DerivativesBaseColumns.Date].values[0]
      return pd.to_datetime(break_even_date)

    def prepare_summary(self):
      self.caclulate_from_df()
      
      if (len(self.non_tradables)) > 0:
        self.non_tradables_no_entry_count = self.non_tradables[self.non_tradables['reason'] == 'No Entry'][DerivativesBaseColumns.Identifier].count()
        self.non_tradables_no_exit_count = self.non_tradables[self.non_tradables['reason'] == 'No Exit'][DerivativesBaseColumns.Identifier].count()

      super().prepare_summary_df()
    
    def save_result(self, excel_writer: pd.ExcelWriter, suffix: str):
      if self.summary is not None:
        self.summary.to_excel(excel_writer, f'{suffix}-summary')
      
      self.trades.to_excel(excel_writer, f'{suffix}-trades')
      self.full_trades.to_excel(excel_writer, f'{suffix}-full-trades')
      self.non_tradables.to_excel(excel_writer, f'{suffix}-non-tradables')

    def load_from_file(file_path: str, suffix: str):
      trades = pd.read_excel(file_path, f'{suffix}-trades')
      full_trades = pd.read_excel(file_path, f'{suffix}-full-trades')
      non_tradables = pd.read_excel(file_path, f'{suffix}-non-tradables')
      instance = TradesResult(trades, full_trades, non_tradables)
      instance.prepare_summary()
      return instance
      
class TradesResultComposite(TradesResult):
  results: dict
  underlying_summaries: pd.DataFrame = None

  def __init__(self):
    super().__init__(pd.DataFrame(), pd.DataFrame(), pd.DataFrame())
    self.results = {}
    self.underlying_summaries = pd.DataFrame()

  def add_result(self, result: TradesResult, key: str):
    self.results[key] = result
    
    if len(result.trades) > 0:
      result.trades['id'] = key
      self.trades = pd.concat([self.trades, result.trades], ignore_index=True)
    
    if len(result.non_tradables) > 0:
      result.non_tradables['id'] = key
      self.non_tradables = pd.concat([self.non_tradables, result.non_tradables], ignore_index=True)

    if len(result.full_trades) > 0:
      result.full_trades['id'] = key
      self.full_trades = pd.concat([self.full_trades, result.full_trades], ignore_index=True)

    if result.summary is not None and len(result.summary) > 0:
      result.summary['id'] = key
      self.underlying_summaries = pd.concat([self.underlying_summaries, result.summary], ignore_index=True)
  
    self.prepare_summary()

  def get_underlying_summaries(self):
    return self.underlying_summaries

  def save_result(self, path: str):
    excel_writer = pd.ExcelWriter(path)
    super().save_result(excel_writer, '')
    self.get_underlying_summaries().to_excel(excel_writer, 'underlying-summaries')
    excel_writer.close()

class BiDirectionalTradesResult(TradesResultComposite):
  bear_trades_result: TradesResult
  bull_trades_result: TradesResult

  def __init__(self, bear_trades_result: TradesResult, bull_trades_result: TradesResult):
    super().__init__()
    self.bear_trades_result = bear_trades_result
    self.bull_trades_result = bull_trades_result
    self.add_result(bear_trades_result, 'bear')
    self.add_result(bull_trades_result, 'bull')

  def load_from_file(base_path: str):
    bear_trades_result = TradesResult.load_from_file(base_path, 'bear')
    bull_trades_result = TradesResult.load_from_file(base_path, 'bull')
    return BiDirectionalTradesResult(bear_trades_result, bull_trades_result)

  def save_result(self, path):
    excel_writer = pd.ExcelWriter(path)
    self.summary.to_excel(excel_writer, f'full-summary')
    
    for key in self.results:
      result = self.results[key]
      result.save_result(excel_writer, key)

    excel_writer.close()


from market_insights.core.column_definition import DerivativesBaseColumns, DerivativesCalculatedColumns
from market_insights.core.core import TypeHelper, MarketDaysHelper
import datetime
import pandas as pd
import numpy as np

from market_insights.trade_builders.results import BiDirectionalTradesResult, TradesResultComposite, TradesResult

class DerivativeTradeOptions:
  holding_days: int = 7
  budget_min: int = 8000
  budget_max: int = 12000
  min_lots_oi: int = 50
  entry_amount_col: str = DerivativesCalculatedColumns.CloseAmount
  exit_amount_profit_col: str = DerivativesCalculatedColumns.HighAmount
  exit_amount_loss_col: str = DerivativesCalculatedColumns.LowAmount
  exit_amount_close_col: str = DerivativesCalculatedColumns.CloseAmount
  profit_factor: float = 2
  loss_factor: float = 0.5
  entry_lag_days: int = 0

class DerivativesNonTradablesDiagnostic:
  class EntryOptions:
    skip_budget_check = False
    skip_expiry_date_check = False
    skip_oi_check = False

  class ExitOptions:
    skip_holding_days_check = False
    skip_oi_check = False

class OptionsTradeBuilder():
  _trade_options: DerivativeTradeOptions
  _option_type: str
  
  def __init__(self, trade_options, option_type: str):
    self._trade_options = trade_options
    self._option_type = option_type

  def get_last_exit_date(self, signal_date):
     return self.get_first_entry_date(signal_date) + datetime.timedelta(days = self._trade_options.holding_days)
  
  def get_first_entry_date(self, signal_date):
    if self._trade_options.entry_lag_days > 0:
      for_date = pd.to_datetime(signal_date)
      for i in range(0, self._trade_options.entry_lag_days):
        for_date = MarketDaysHelper.get_next_market_day(for_date)
      return pd.to_datetime(for_date)
    else:
      return pd.to_datetime(signal_date)
  
  def get_full_trade(self, buy_option, sell_option, index):
    return pd.DataFrame({
        DerivativesBaseColumns.Identifier: buy_option[DerivativesBaseColumns.Identifier].values[0],
        DerivativesBaseColumns.StrikePrice: buy_option[DerivativesBaseColumns.StrikePrice].values[0],
        DerivativesBaseColumns.ExpiryDate: buy_option[DerivativesBaseColumns.ExpiryDate].values[0],
        DerivativesBaseColumns.OptionType: buy_option[DerivativesBaseColumns.OptionType].values[0],
        'EntryDate': buy_option[DerivativesBaseColumns.Date].values[0],
        'Debit': -1 * buy_option['Credit'].values[0],
        'ExitDate': sell_option[DerivativesBaseColumns.Date].values[0],
        'Credit': sell_option['Credit'].values[0],
        'Quantity': 1 * buy_option['Quantity'].values[0],
        'Turnover': abs(buy_option['Credit'].values[0]) + abs(sell_option['Credit'].values[0]),
        'Net': buy_option['Credit'].values[0] + sell_option['Credit'].values[0]
    }, index=index)

  def get_entry_option(self, data, symbol, signal_date, index = 0, diagnostic = DerivativesNonTradablesDiagnostic.EntryOptions()):
    hold_till_date = self.get_last_exit_date(signal_date)
    entry_option = data[
        (data[DerivativesBaseColumns.Identifier] == symbol)
        & (data[DerivativesBaseColumns.OptionType] == self._option_type)
        & (data[DerivativesBaseColumns.Date] == self.get_first_entry_date(signal_date))
        & (
            (data[self._trade_options.entry_amount_col].between(self._trade_options.budget_min, self._trade_options.budget_max))
            |
            (diagnostic.skip_budget_check == True)
          )
        & (
            (pd.to_datetime(data[DerivativesBaseColumns.ExpiryDate]) >= hold_till_date)
            |
            (diagnostic.skip_expiry_date_check == True)
          )
        & (
            (data[DerivativesBaseColumns.OpenInterest] >= (data[DerivativesCalculatedColumns.LotSize] * self._trade_options.min_lots_oi))
            |
            (diagnostic.skip_oi_check == True)
          )
    ].sort_values(DerivativesBaseColumns.StrikePrice, ascending=False).head(1)

    if len(entry_option) > 0:
        strike_price = entry_option[DerivativesBaseColumns.StrikePrice].values[0]
        expiry_date = entry_option[DerivativesBaseColumns.ExpiryDate].values[0]
        buy_amount = entry_option[self._trade_options.entry_amount_col].values[0]
        lotsize = entry_option[DerivativesCalculatedColumns.LotSize].values[0]
        entry_date = entry_option[DerivativesBaseColumns.Date].values[0]

        buy = pd.DataFrame({
                    DerivativesBaseColumns.Identifier: symbol,
                    DerivativesBaseColumns.StrikePrice: strike_price,
                    DerivativesBaseColumns.ExpiryDate: expiry_date,
                    DerivativesBaseColumns.OptionType: self._option_type,
                    DerivativesBaseColumns.Date: entry_date,
                    'Credit': -1 * buy_amount,
                    'Quantity': 1 * lotsize,
                    'Type': 'Buy'
                }, index=[(index*2)])
        return buy
    else:
        return []
    
  def get_exit_option_for_buy_option(self, data, buy_option, index = 0, diagnostic = DerivativesNonTradablesDiagnostic.ExitOptions()):
    return self.get_exit_option(
        data,
        symbol = buy_option[DerivativesBaseColumns.Identifier].values[0],
        expiry_date = buy_option[DerivativesBaseColumns.ExpiryDate].values[0],
        trade_date = buy_option[DerivativesBaseColumns.Date].values[0],
        strike_price = buy_option[DerivativesBaseColumns.StrikePrice].values[0],
        lotsize = buy_option['Quantity'].values[0],
        buy_amount = abs(buy_option['Credit'].values[0]),
        index = index,
        diagnostic = diagnostic
      )
      
  def get_exit_option(self, data, symbol, expiry_date, trade_date, strike_price, lotsize, buy_amount, index = 0, diagnostic = DerivativesNonTradablesDiagnostic.ExitOptions()):
    hold_till_date = self.get_last_exit_date(trade_date)
    exit_options = data[
        (data[DerivativesBaseColumns.Identifier] == symbol)
        & (data[DerivativesBaseColumns.ExpiryDate] == expiry_date)
        & (data[DerivativesBaseColumns.StrikePrice] == strike_price)
        & (data[DerivativesBaseColumns.OptionType] == self._option_type)
        & (
            (data[DerivativesBaseColumns.Date].between(pd.to_datetime(trade_date) + datetime.timedelta(days=1), hold_till_date))
            |
            (diagnostic.skip_holding_days_check == True)
          )
        & (
            (data[DerivativesBaseColumns.OpenInterest] >= (data[DerivativesCalculatedColumns.LotSize] * self._trade_options.min_lots_oi))
            |
            (diagnostic.skip_oi_check == True)
          )
    ].sort_values(DerivativesBaseColumns.Date)

    if len(exit_options) > 0:
        exit_option = exit_options[(
            (exit_options[self._trade_options.exit_amount_profit_col] >= (buy_amount * self._trade_options.profit_factor))
            |
            (exit_options[self._trade_options.exit_amount_loss_col] <= (buy_amount * self._trade_options.loss_factor))
          )]
        
        sell_amount = 0

        if len(exit_option) <= 0:
            exit_option = exit_options.sort_values(DerivativesBaseColumns.Date, ascending=False)[0:1]
            sell_amount = exit_option[self._trade_options.exit_amount_close_col].values[0]
        else:
          high_amount = exit_option[self._trade_options.exit_amount_profit_col].values[0]
          low_amount = exit_option[self._trade_options.exit_amount_loss_col].values[0]
          
          if (high_amount >= (buy_amount * self._trade_options.profit_factor)):
            sell_amount = (buy_amount * self._trade_options.profit_factor)
          else:
            sell_amount = min(buy_amount * self._trade_options.loss_factor, high_amount)

        #if (high_amount >= (buy_amount * self._trade_options.profit_factor)):
        #    sell_amount = (buy_amount * self._trade_options.profit_factor)
        #else:
        #    sell_amount = (buy_amount * self._trade_options.loss_factor)

        sell = pd.DataFrame({
            DerivativesBaseColumns.Identifier: symbol,
            DerivativesBaseColumns.StrikePrice: strike_price,
            DerivativesBaseColumns.ExpiryDate: expiry_date,
            DerivativesBaseColumns.OptionType: self._option_type,
            DerivativesBaseColumns.Date: pd.to_datetime(exit_option[DerivativesBaseColumns.Date].values[0]),
            'Credit': sell_amount,
            'Quantity': -1 * lotsize,
            'Type': 'Sell'
        }, index=[(index*2)+1])
        return sell
    else:
        return []

class BearOptionsTradeBuilder(OptionsTradeBuilder):
  def __init__(self, trade_options):
     super().__init__(trade_options, 'PE')

class BullOptionsTradeBuilder(OptionsTradeBuilder):
  def __init__(self, trade_options):
     super().__init__(trade_options, 'CE')

class DerivatesMultiTradeOptions:
  avoid_overlap_trades: bool = True

class OptionsMultiTradesBuilder:
    def __init__(self, trade_options: DerivatesMultiTradeOptions):
        self.trade_options = trade_options

    def get_trades(self, symbol_signals, options_data, trade_builder: OptionsTradeBuilder):
        trades: pd.DataFrame = pd.DataFrame()
        full_trades: pd.DataFrame = pd.DataFrame()
        non_tradables: pd.DataFrame() = pd.DataFrame()

        symbols = symbol_signals[DerivativesBaseColumns.Identifier].unique()
        for symbol in symbols:
            symbol_options_data = options_data[options_data[DerivativesBaseColumns.Identifier] == symbol]
            for index, symbol_signal in symbol_signals[symbol_signals[DerivativesBaseColumns.Identifier] == symbol].iterrows():
                signal_date = pd.to_datetime(symbol_signal[DerivativesBaseColumns.Date])
                if len(trades) > 0:
                  symbol_outstanding_trade_quantity = trades[
                     (trades[DerivativesBaseColumns.Identifier] == symbol)
                     &
                     (trades[DerivativesBaseColumns.Date] <= signal_date)
                    ]['Quantity'].sum()
                else:
                   symbol_outstanding_trade_quantity = 0

                if (symbol_outstanding_trade_quantity == 0 or self.trade_options.avoid_overlap_trades == False):
                  symbol_options_data_for_dates = symbol_options_data[symbol_options_data[DerivativesBaseColumns.Date].between(signal_date, trade_builder.get_last_exit_date(signal_date))]
                  
                  buy_option = trade_builder.get_entry_option(symbol_options_data_for_dates, symbol, signal_date, index * 2)
                  
                  if len(buy_option) > 0:
                      sell_option = trade_builder.get_exit_option_for_buy_option(symbol_options_data_for_dates, buy_option, (index * 2) + 1)
                      
                      if len(sell_option) > 0:
                          trades = pd.concat([trades, buy_option, sell_option], ignore_index=True)
                          full_trades = pd.concat([full_trades, trade_builder.get_full_trade(buy_option, sell_option, index=[(index*2)])], ignore_index=True)
                      else:
                          non_tradables = pd.concat([non_tradables, pd.DataFrame({DerivativesBaseColumns.Identifier: symbol, DerivativesBaseColumns.Date: pd.to_datetime(signal_date), DerivativesBaseColumns.OptionType: trade_builder._option_type, 'reason': 'No Exit'}, index=[(index*2)])])
                  else:
                      non_tradables = pd.concat([non_tradables, pd.DataFrame({DerivativesBaseColumns.Identifier: symbol, DerivativesBaseColumns.Date: pd.to_datetime(signal_date), DerivativesBaseColumns.OptionType: trade_builder._option_type, 'reason': 'No Entry'}, index=[(index*2)+1])])
        
        result = TradesResult(trades, full_trades, non_tradables)
        result.prepare_summary()

        return result
    
class StrategyAnalyser:
    multi_trade_options: dict = {}
    trade_options: dict = {}
    results: TradesResultComposite
    output_folder: str = ''

    def __init__(self, output_sub_folder: str):
        self.results = TradesResultComposite()
        self.multi_trade_options = {}
        self.trade_options = {}
        self.output_folder = '../output/' + output_sub_folder
        
    def set_multi_trade_options(self, options: dict):
        self.multi_trade_options = options
        return self
    
    def set_trade_options(self, options: dict):
        self.trade_options = options
        return self

    def iterator(self, callback):
        for key1 in self.multi_trade_options:
            for key2 in self.trade_options:
              callback(key1, key2, self.multi_trade_options[key1], self.trade_options[key2])

    def load_from_files(self):
        for key1 in self.multi_trade_options:
            for key2 in self.trade_options:    
                result = BiDirectionalTradesResult.load_from_file(f'{self.output_folder}/{key1}-{key2}.xlsx')
                self.results.add_result(result, f'{key1}-{key2}')
        return self.results

    def analyze(self, options_data: pd.DataFrame, sell_signal_data_file: str = 'sell_signal_symbols.csv', buy_signal_data_file: str = 'buy_signal_symbols.csv'):
        sell_signal_data = pd.read_csv(f'{self.output_folder}/{sell_signal_data_file}')
        buy_signal_data = pd.read_csv(f'{self.output_folder}/{buy_signal_data_file}')

        sell_signal_data[DerivativesBaseColumns.Date] = pd.to_datetime(sell_signal_data[DerivativesBaseColumns.Date])
        buy_signal_data[DerivativesBaseColumns.Date] = pd.to_datetime(buy_signal_data[DerivativesBaseColumns.Date])

        sell_signal_symbols = sell_signal_data[DerivativesBaseColumns.Identifier].unique()
        buy_signal_symbols = buy_signal_data[DerivativesBaseColumns.Identifier].unique()

        all_symbols = np.unique(np.concatenate((sell_signal_symbols, buy_signal_symbols)))

        options_data = options_data[options_data[DerivativesBaseColumns.Identifier].isin(all_symbols)]

        self.results = TradesResultComposite()
        for key1 in self.multi_trade_options:
            multi_trade_builder = OptionsMultiTradesBuilder(self.multi_trade_options[key1])
            for key2 in self.trade_options:
                bear_trade_builder = BearOptionsTradeBuilder(self.trade_options[key2])
                bull_trade_builder = BullOptionsTradeBuilder(self.trade_options[key2])
                bear_trades = multi_trade_builder.get_trades(sell_signal_data, options_data, bear_trade_builder)
                bull_trades = multi_trade_builder.get_trades(buy_signal_data, options_data, bull_trade_builder)
                
                result = BiDirectionalTradesResult(bear_trades, bull_trades)
                result.save_result(f'{self.output_folder}/{key1}-{key2}.xlsx')
                self.results.add_result(result, f'{key1}-{key2}')
        #self.results.save_result(f'{self.output_folder}/final.xlsx')
        return self.results


class EnvironmentSettings:
    Paths = {
        'DataBaseDir': "../../_data",
        'RawDataDir': "raw",
        'ProcessedDataDir': "processed",
        'DailyDataDir': "daily",
        'HistoricalDataDir': "historical",
        'MonthlySuffix': "monthly",
        'AnnualSuffix': "annual",
        'BhavDataDir': "bhavcopy",
        'NseDerivativesDataDir': "nse_derivatives",
        'NseIndicesDataDir': "nse_indices",
        'ManualDataPath': "../manual_data"
    }
    Development = {
        "InstrumentationLevel": 1 | 2 | 4
    }
